# %load LabQuiz.old/__init__.py

from importlib.metadata import version, PackageNotFoundError
from .i18n import set_language, _

try:
    __version__ = version("labquiz")
except PackageNotFoundError:
    __version__ = "0.0.0"

from .utils import StudentForm
from .main import QuizLab
__all__ = [
    QuizLab,
    StudentForm,
    set_language
]



import yaml
import random
import sys, io
from io import StringIO
from pathlib import Path
import copy
import re
import numpy as np
import pandas as pd
import requests
import ipywidgets as widgets
from IPython.display import display, HTML
from datetime import datetime
import hashlib, base64, json
import tarfile
from cryptography.fernet import Fernet, InvalidToken

qz = ('prep '*3).strip().split(' ')
qz.extend(['_get_protected_data', 'show', 'QuizLab'])

def shuffle_quiz_propositions(input_file, output_file):
    # Reading the YAML file
    with open(input_file, "r", encoding="utf-8") as f:
        data = yaml.safe_load(f)

    # Shuffle the order of propositions for each quiz
    for quiz_name, quiz_content in data.items():
        if "propositions" in quiz_content:
            random.shuffle(quiz_content["propositions"])

    # Saving in a new file
    with open(output_file, "w", encoding="utf-8") as f:
        yaml.safe_dump(
            data,
            f,
            allow_unicode=True,
            sort_keys=False
        )


def encode_dict_base64(obj: dict) -> str:
    json_str = json.dumps(obj, ensure_ascii=False)
    return base64.b64encode(json_str.encode("utf-8")).decode("ascii")
    

def encode_file(input_file, output_file):
    with open(input_file, "r", encoding="utf-8") as f:
        data = yaml.safe_load(f)

    for quiz in data.values():
        if "propositions" in quiz:
            quiz["propositions"] = [
                encode_dict_base64(p) for p in quiz["propositions"]
            ]
        if "constraints" in quiz:
            quiz["constraints"] = [
                encode_dict_base64(p) for p in quiz["constraints"]
            ]

    with open(output_file, "w", encoding="utf-8") as f:
        yaml.safe_dump(data, f, allow_unicode=True, sort_keys=False)
        
def crypt_file(input_file, output_file, pwd='', verbose=False):
    
    ctx = qz + [output_file, pwd]
     
    with open(input_file, "r", encoding="utf-8") as f:
        data = yaml.safe_load(f)
        
    if "title" in data.keys(): 
        key = b'zzFJ0WBRXuFC1qJOFGxeqt3NUNcFR9vdZqFkCry-DYw='
        frt = Fernet(key)
        data['title'] = frt.encrypt(data['title'].encode("utf-8") )
    
    for quiz_id, quiz in data.items():
        if quiz_id == "title": continue
        nctx = ctx + [quiz_id]
        fingerprint = "-".join(nctx).encode()
        #print("fingerprint", fingerprint)
        hash_digest = hashlib.sha256(fingerprint).digest()
        key = base64.urlsafe_b64encode(hash_digest)
        frt = Fernet(key)
        data[quiz_id] = frt.encrypt(json.dumps(quiz).encode("utf-8") ) 
        if verbose: print(quiz)
               

    with open(output_file, "w", encoding="utf-8") as f:
        yaml.safe_dump(data, f, allow_unicode=True, sort_keys=False)


def decrypt_file(input_file, output_file=None, pwd='', verbose=False):
    
    ctx = qz + [output_file, pwd]
    #ctx = ['prep', 'prep', 'prep', '_get_protected_data', 'show', 'QuizLab']
    fingerprint = "-".join(ctx).encode()
    hash_digest = hashlib.sha256(fingerprint).digest()
    key = base64.urlsafe_b64encode(hash_digest)
    frt = Fernet(key)
    
    with open(input_file, "r", encoding="utf-8") as f:
        data = yaml.safe_load(f)

    if "title" in data.keys(): 
        key = b'zzFJ0WBRXuFC1qJOFGxeqt3NUNcFR9vdZqFkCry-DYw='
        frt = Fernet(key)
        data['title'] = frt.decrypt(data['title'].decode("utf-8") )
         
    for quiz_id, quiz in data.items():
        if quiz_id == "title": continue
        nctx = ctx + [quiz_id]
        fingerprint = "-".join(nctx).encode()
        #print("fingerprint", fingerprint)
        hash_digest = hashlib.sha256(fingerprint).digest()
        key = base64.urlsafe_b64encode(hash_digest)
        #print("key", key)
        frt = Fernet(key)
        
        quiz = json.loads(frt.decrypt(quiz).decode('utf-8')) 
        data[quiz_id] = quiz

        if verbose: 
            print(quiz)
            
    if output_file is not None:
        with open(output_file, "w", encoding="utf-8") as f:
             yaml.safe_dump(data, f, allow_unicode=True, sort_keys=False)
#decrypt_file("quizzes_basic_filtering_test.enc", verbose=True)
        
        
def quiz_propositions_for_exam(input_file, output_file):
    # Input: initial or mixed question file 
    # # Output: File without solutions, answers or tips 
    # # Reading the YAML file
    with open(input_file, "r", encoding="utf-8") as f:
        data = yaml.safe_load(f)

    # removing hints for each quiz
    for quiz_name, quiz_content in data.items():
        if quiz_name == "title": continue 
        quiz_content.pop("constraints", None)
        for prop in quiz_content["propositions"]:
            '''
            prop["expected"] = ""
            prop["reponse"] = ""
            prop["tip"] = ""
            '''  
            keys_to_remove = {"expected", "reponse", "tip"}
            for k in keys_to_remove:
                prop.pop(k, None)
            
    # Saving in a new file
    with open(output_file, "w", encoding="utf-8") as f:
        yaml.safe_dump(
            data,
            f,
            allow_unicode=True,
            sort_keys=False
        )


def quiz_to_dict_of_solutions(input_file):
    """
    extracts the solutions from the initial file

    Settings
    ----------
        input_file:str
            original or mixed question file    
    Returns
    -------
        out: dict
            dictionary {quiz1: {label1:expected, label2:expected, ...}, 
            quiz2:{label1:expected, label2:expected, ...} }
    """
    # Reading the YAML file
    with open(input_file, "r", encoding="utf-8") as f:
        data = yaml.safe_load(f)

    # 
    dico = {quiz_name:  {prop["label"]:prop["expected"] for prop in quiz_content["propositions"]} 
            for quiz_name, quiz_content in data.items()}
    out = {l: dict(sorted(d.items(), key=lambda item: item[0]))
                       for l,d in zip(dico.keys(), dico.values())}
    # We sort on the keys in case of!
    
    return out
        

def crypt_data(data, output_file, pwd=''):
    
    data = copy.deepcopy(data)
    ctx = qz + [output_file, pwd]
    if "title" in data.keys(): 
        key = b'zzFJ0WBRXuFC1qJOFGxeqt3NUNcFR9vdZqFkCry-DYw='
        frt = Fernet(key)
        data['title'] = frt.encrypt(data['title'].encode("utf-8") )
    
    for quiz_id, quiz in data.items():
        if quiz_id == "title": continue
        nctx = ctx + [quiz_id]
        fingerprint = "-".join(nctx).encode()
        hash_digest = hashlib.sha256(fingerprint).digest()
        key = base64.urlsafe_b64encode(hash_digest)
        frt = Fernet(key)
        data[quiz_id] = frt.encrypt(json.dumps(quiz).encode("utf-8") ) 
    return data


def encode_data(data):
    data = copy.deepcopy(data)
    for quiz in data.values():
        if "propositions" in quiz:
            quiz["propositions"] = [
                encode_dict_base64(p) for p in quiz["propositions"]
            ]
        if "constraints" in quiz:
            quiz["constraints"] = [
                encode_dict_base64(p) for p in quiz["constraints"]
            ]
    return data


def prepare_files(input_file, output_file, mode="crypt", pwd=""):    
    """
    Prepare YAML files for quizzes. 
    Outputs two files, with the basename given in `output_file`. 
    The second file is questions only and is the input stripped
    from responses ans tips. With the `mode="crypt"`, the input and stripped 
    versions are encrypted; with the `mode="enc"`, both files are binhex encoded; 
    finally, with `mode=yml`, files are not encoded nor encrypted.

    Parameters
    ----------
    input_file : str
        Path to the input YAML file.
    output_file : str
        Path to the output YAML file.
    mode : str, optional
        Mode to prepare the file. Choose from "crypt", "enc", or "yml".
    pwd : str, optional
        Password for file encryption.

    Returns
    -------
    None
    """    
    import copy
    # Lecture du fichier YAML
    with open(input_file, "r", encoding="utf-8") as f:
        data = yaml.safe_load(f)

    # Shuffle the order of propositions in each quiz
    for quiz_name, quiz_content in data.items():
        if "propositions" in quiz_content:
            random.shuffle(quiz_content["propositions"])
            
    # Questions only
    data_only = copy.deepcopy(data)
    # removing hints for every quiz
    for quiz_name, quiz_content in data_only.items():
        if quiz_name == "title": continue 
        quiz_content.pop("constraints", None)
        for prop in quiz_content["propositions"]:
            keys_to_remove = {"expected", "reponse", "tip"}
            for k in keys_to_remove:
                prop.pop(k, None)
  
    path = Path(output_file)
    if mode == "crypt":
        data_out = crypt_data(data, f"{path.stem}_crypt.txt", pwd=pwd)
        data_only_out = crypt_data(data_only, f"{path.stem}_qo_crypt.txt",  pwd=pwd)
    elif mode == "enc":
        data_out = encode_data(data)
        data_only_out = encode_data(data_only) 
    else:
        mode = "yml"
        data_out = data
        data_only_out = data_only
        
    for suffix, data in zip([mode, f"qo_{mode}"], [data_out, data_only_out]):
        out = path.with_name(f"{path.stem}_{suffix}.txt")
        
        with out.open("w", encoding="utf-8") as f:
            print(f"- Creating {out}")
            yaml.safe_dump(
                data,
                f,
                allow_unicode=True,
                sort_keys=False
            )
    if mode == "crypt" and pwd != '':
        print("⚠️ File crypted with pwd. Ensure to use the `madatoryInternet=True` option in quiz init")    
    
    
#=====================================================
# Lecture des données et correction
#=====================================================

def readData(URL, SECRET):
    """Reading data

    readData(URL, SECRET)
    return df, df_last
    Connects to retrieve the data table designated by the URL address and protected by SECRET

    Settings
    ----------
        URL: str
                Path
        SECRET: str
                String used to access data
    Returns
    -------
        df: pandas dataframe
                complete table
        df_filt:pandas dataframe
                table filtered just on validations and corrections
    """
    
    #global df, df_last

    try:
        r = requests.get(URL, params={"secret": SECRET})
        r.raise_for_status()
        df = pd.read_csv(StringIO(r.text)) 
        df.student = df.student.apply(lambda s: s.strip() if isinstance(s, str) else s)  #Normalization of student column
        df["answers"] = df.answers.apply(lambda x: parse_custom_dict(x)) #Decoding answers column
        df["parameters"] = df.parameters.apply(lambda x: parse_custom_dict(x)) #Decoding answers column
        now = datetime.now().strftime("%H:%M:%S")

        footer_msg = f"<span style='color:green; font-weight:bold'>✔ Sheet loaded ... ({len(df)} rows) - {now} </span>"
        display(HTML(footer_msg))

    except Exception as e:
        footer_msg =  f"<span style='color:red; font-weight:bold'>✘ Loading error {e}</span>"
        display(HTML(footer_msg))
        
    df = df.sort_values("timestamp")
    #df_last = df.drop_duplicates(
    #    subset=["student", "quiz_title", "event_type"],
    #    keep="last"
    #)
    #df_last_validate or correction 
    #df_last = df_last.query("event_type == 'validate' ")
    df_last = df.query("event_type in ['validate', 'validate_exam', 'correction'] ")
    return df, df_last

def convert_ans(answers):
    """Conversion of answers (str) to dictionary

    convert_ans(answers)
    return dictionary
    Converts responses (string) to response dictionary

    Settings
    ----------
        answers: str
    Returns
    -------
        dico: dict
                    matching dictionary
    """

    try:
        dico = {k: v.lower() == "true" for k, v in (item.rsplit("=",1) for item in answers.strip("{}").split(", "))}
        dico = dict(sorted(dico.items(), key=lambda item: item[0])) # We sort keys, never mind
    except:
        print(answers)
    return dico


def correct_ans(quiz, quiz_id, given, weights=None, constraints=None):
    """
    Correction of an answer
    # correct_ans(quiz, quiz_id, given, weights=None)
    # return score, score_max
    Correct the answers given in `given` by comparing with those contained
    in the subject (with correction) for the label `quiz_id`. 

    Settings
    ----------
        quiz: object 
            QuizLab object into which the topic was loaded
            >> quiz = QuizLab(URL, QUIZFILE, needAuthentification=False, retries=100)
        quiz_id: str
            Key to the desired quiz
        given: dict 
            Answers given
        weights: pandas dataframe
            Weight matrix 
            ex: default_weights = {
                (True, True): 1, # True Positive
                (True, False): -1, # False Positive 
                (False, True): 0, # False Negative (forget)
                (False, False): 0 # True Negative
                }
    Returns
    -------
        score:float
                score obtained
        score_max: float
                maximum score that can be achieved
    """

    from .utils import calculate_quiz_score
    question, quiz_type, propositions, constraints = quiz._QuizLab__load_quiz(quiz_id)
    #NB: The name _QuizLab__load_quiz corresponds to the dunder __loadquiz method in main
    propositions.sort(key=lambda d: d["label"]) # tri sur les clés
    given = {k:given[k] for k in sorted(given)}
    #print(given, propositions)
    return calculate_quiz_score(quiz_type, given, propositions, weights=weights, constraints=constraints)


def check_ans(student, quiz_id, df, maxtries=3):
    """Selects the response to use

    check_ans(student, quiz_id, df, maxtries=maxtries)
    return given
    Select the “correct” answer: 
    - that of rank <= maxtries 
    - if correction requested: the last before correction if <= maxtries otherwise maxtries 

    Arguments
    ---------
        student:str
            student name to search in the table
        quiz_id: str
            id of the string to use (to search in the table)
        df: pandas dataframe
            data table
        maxtries: int
            Max number of attempts
    Returns
    -------
        given: dict
            answer dictionary
    """

    extract = df.query(f"student == '{student}' & quiz_title=='{quiz_id}'").reset_index()
    if len(extract):
        correction = extract[extract['event_type'] == 'correction'].index
        if len(correction):
            idx = (correction[0]-1) if (correction[0]-1) <= maxtries else maxtries
        else: 
            idx = len(extract)-1 if (len(extract)<=maxtries) else maxtries
        
        if (idx==-1): 
            #print(quiz_id, "A demandé la correction avant valid")
            return {}
        #quiz_test = extract.loc[idx]["answers"]
        #given = convert_ans(quiz_test)
        given = extract.loc[idx]["answers"]
        #print(extract.loc[idx]["answers"])
    else:
        given = {}
    return given   



def new_parse_custom_dict(text):  # Format change!! Much simpler
    """Converts the character string answers into a dictionary

    parse_custom_dict(text)
    return result
    With format change, just json decoding

    Settings
    ----------
        text:str
            character string in the format '{"val1":true, "val2":false, "val3":float, "val4":"{...}"}'
    Returns
    -------
        result: dict
            sorted dictionary of identified keys:values       
    """

    import json
    d = json.loads(text)
    if isinstance(d, dict):
        d_trie = {k: v for k, v in sorted(d.items())}
        return d_trie
    return d

def old_parse_custom_dict(text):
    """Converts a string into a dictionary

    parse_custom_dict(text)
    return result
    Converts a character string to the format "{val1=true, val2=false, val3=float, val4={...}}"
    to a dictionary containing boolean values, numbers, or other dictionaries

    Settings
    ----------
        text:str
            character string in the format "{val1:true, val2=false, val3=float, val4={...}}"
    Returns
    -------
        result: dict
            dictionary of keys:identified values

    #Thanks to Gemini!!         
    """
    # Nettoyage des accolades extérieures
    text = text.strip().strip("{}")
    result = {}
    
    # 1. Découpage intelligent par virgule (ignore les virgules dans les { })
    parts = []
    bracket_level = 0
    current = []
    for char in text:
        if char == '{': bracket_level += 1
        elif char == '}': bracket_level -= 1
        
        if char == ',' and bracket_level == 0:
            parts.append("".join(current).strip())
            current = []
        else:
            current.append(char)
    if current:
        parts.append("".join(current).strip())

    # 2. Analyse de chaque partie
    for part in parts:
        if not part: continue
        
        # On cherche le DERNIER signe '=' qui sépare la clé de la valeur.
        # La valeur peut être :
        # - Un bloc entre accolades { ... }
        # - Une chaîne (hash, texte, booléen, nombre)
        if '=' in part:
            # On découpe par la droite pour gérer "Sinc x=0 Handling=false"
            # On cherche le '=' qui précède une valeur type (bool, num, hash, ou {)
            match = re.search(r'^(.*)=(.*)$', part)
            if match:
                key = match.group(1).strip()
                val_raw = match.group(2).strip()
                
                # Conversion des types
                if val_raw.startswith('{'):
                    result[key] = parse_custom_dict(val_raw)
                elif val_raw.lower() == 'true':
                    result[key] = True
                elif val_raw.lower() == 'false':
                    result[key] = False
                else:
                    try:
                        # On tente le nombre, sinon on garde la chaîne (hash, etc.)
                        if '.' in val_raw:
                            result[key] = float(val_raw)
                        else:
                            result[key] = int(val_raw)
                    except ValueError:
                        result[key] = val_raw # C'est ici que les hash sont sauvés
                        
        result = dict(sorted(result.items(), key=lambda item: item[0])) # On trie les clés, on ne sait jamais
                         
    return result

def parse_custom_dict(text):
    """Converts string of answers in df['answers'] to dictionary, keys sorted
    """
    if pd.isnull(text): return text
    try: 
        result = new_parse_custom_dict(text)
    except Exception as e:
        #print(f"Erreur level 1 décodage parse_custom_dict() pour {text}")
        #print("Erreur level 1", e)
        try:
            result = old_parse_custom_dict(text)
        except Exception as e:
            print(f"Erreur level 2 décodage parse_custom_dict() pour {text}")
            print("Erreur level 2", e)
            return text
    return result



def correctAll(students_answers, quiz, df, seuil=0, exam_questions=None, weights=None, bareme=None, maxtries=3):
    """Correct all answers

    correctAll(students, floors, df, threshold=0, Weight=None, scale=None)
    return Res 
    Corrects all student results from `students` using `sols` solutions

    Arguments
    ---------
        students: dict
            Answers given
        quiz: object 
            QuizLab object into which the topic was loaded
            >> quiz = QuizLab(URL, QUIZFILE, needAuthentification=False, retries=100)
        df: pandas dataframe
            Input table
        threshold: float
            threshold=0 thresholds the marks for each question to zero (otherwise negative mark possible)
        exam_questions: list (default None)
            list of questions to correct (by default all)
        weights: dict
            Weight matrix 
            ex: default_weights = {
                (True, True): 1, # True Positive
                (True, False): -1, # False Positive 
                (False, True): 0, # False Negative (forget)
                (False, False): 0 # True Negative
            }
        scale: dict
            Scale = weight of the different questions in the quiz. If no scale, all 
            questions have the same weight for calculating the grade. If weight of a question 
            not specified, it is 1 by default. ex: scale = {'quiz3':4, 'quiz55':0} (all others 
            with a weight of 1).
        maxtries: int
            Number of attempts allowed
        
    Returns
    -------
        Res: pandas dataframe
            rows: students, columns: each quiz + a Note column 

    """
    import pandas as pd
    
    allquizzes = [q for q in quiz.quiz_bank.keys() if q != "title"]
    students = students_answers.keys()
    if exam_questions is None: exam_questions = {s:allquizzes for s in students}
    Res = pd.DataFrame(index=students, columns=allquizzes)
    if bareme is not None: 
        bareme = {q:1 if q not in bareme.keys() else bareme[q] for q in allquizzes }
    else:
        bareme = {q:1 for q in allquizzes}
    for s in students:
        for quiz_id in exam_questions[s]:
            given = students_answers[s].get(quiz_id, {})
            if given == {}: continue
            score, score_max = 0,1   # précaution            
            try:
                score, score_max = correct_ans(quiz, quiz_id, given, weights=weights)
            except Exception as e:
                print(f"Erreur pour la correction de {s}, quiz_id={quiz_id}")
                #print(given, type(given), e)
            #print(f"Correction de {s}, quiz_id={quiz_id}")
            Res.loc[s, quiz_id] = score/score_max
        Res.loc[s, 'maxpts'] = pd.Series([bareme[q] for q in exam_questions[s]]).sum()
    Res = Res.infer_objects(copy=False).dropna(axis=1, how="all").fillna(0)
    if seuil==0: Res[Res < 0] = 0
    poids = pd.Series({q: bareme[q] for q in Res.columns if q != 'maxpts'})
    Res['Note'] = Res[poids.index].dot(poids) / Res['maxpts'] * 20
      
    return Res

#correctAllBis(students_answers, quiz, data_filt, seuil=0, exam_questions=exam_questions, weights=None, bareme=None, maxtries=3)
# Res = correctAllBis(students_answers, quiz, data_filt, seuil=0, exam_questions=None, weights=None, bareme=None, maxtries=3)
"""
# All students case
import time
tic = time.perf_counter()
ResDataFiltBis = correctAllBis(students_answers, quiz, 
                            data_filt, bareme={ 'quiz13':0, 'quiz55':0, 'quiz56':0, 'quiz57':0})
toc = time.perf_counter()
print(f"Temps d'exécution : {toc-tic:.3f} seconde(s)")
"""

"""
#Random multiple exams
#quiz.exam_show(exam_title = "Examen de Truc",   shuffle=True, nb=2) # <-- the exam itself
exam_questions = getExamQuestions("Examen de Truc", data)
students = exam_questions.keys()
students_answers = getAllStudentsAnsvers(students, data, maxtries=3)
correctAllBis(students_answers, quiz, data_filt, seuil=0, exam_questions=exam_questions, weights=None, bareme=None, maxtries=3)
"""

"""
## 1 - Lecture des données
URL = "https://URL_UTILISÉE_POUR_RECUEILLIR_LES_RÉSULTATS"
SECRET = "MOT_DE_PASSE_SECRET_SPÉCIFIÉ_DANS_LE_SHEET"
QUIZFILE = "quizzes_basic_filtering_test.yml"
"""

def correctQuizzes(URL, SECRET, QUIZFILE, title=None, seuil=0, weights=None, bareme=None, maxtries=1):
    """
    Correct all quizzes in URL
    Arguments
    ---------    
        URL:str
             Address of the Google sheet on which to fetch the data
        SECRET: str
             SECRET code used to access data
        QUIZFILE:str
             Unencoded quiz file name CONTAINING expected values 
        title: str None by default
            If title is not None, it is because we must correct a test with random selection of questions
            of type exam_show, of title title
        threshold: float
            threshold=0 thresholds the marks for each question to zero (otherwise negative mark possible)
        weights: dict
            Weight matrix 
            ex: default_weights = {
                (True, True): 1, # True Positive
                (True, False): -1, # False Positive 
                (False, True): 0, # False Negative (forget)
                (False, False): 0 # True Negative
            }
        scale: dict
            Scale = weight of the different questions in the quiz. If no scale, all 
            questions have the same weight for calculating the grade. If weight of a question 
            not specified, it is 1 by default. ex: scale = {'quiz3':4, 'quiz55':0} (all others 
            with a weight of 1).
        maxtries: int
            Number of attempts allowed
        
    Returns
    -------
        Res: pandas dataframe
            rows: students, columns: each quiz + a Note column 
            
    """
    from labquiz import QuizLab
    from labquiz.putils import readData, getAllStudentsAnsvers, getExamQuestions, correctAll

    ## 1 - Lecture des données
    data, data_filt = readData(URL, SECRET)
    
    if title is None:
        ## 2 - identification des participants et extraction des réponses
        students = sorted(list(data_filt["student"].dropna().unique()))
        exam_questions = None
    else:
        exam_questions = getExamQuestions(title, data)
        students = exam_questions.keys()
    
    students_answers = getAllStudentsAnsvers(students, data, maxtries=maxtries) 

    ## 3 - Instancier un quiz avec le fichier de quiz CONTENANT les valeurs attendues
    URL = ""
    quiz = QuizLab(URL, QUIZFILE, needAuthentification=False, retries=0)

    ## 4 - Corriger !
    ResDataFilt = correctAll(students_answers, quiz, data_filt, seuil=seuil, 
                    exam_questions=exam_questions, weights=weights, bareme=bareme, maxtries=maxtries)
    
    if bareme is not None:
        colsToDrop = [quiz_id for quiz_id, val in bareme.items() if val==0 and quiz_id in ResDataFilt.columns]
        return ResDataFilt.drop(columns=colsToDrop)
    return ResDataFilt


def correctQuizzesDf(data, data_filt, quiz, title=None, seuil=0, weights=None, bareme=None, maxtries=1):
    """
    Correct all quizzes in URL
    Arguments
    ---------    
        data, data_filt: pandas dataframes
             The two tables from readData
        quiz: str
             quiz instance with non-encoded quiz file CONTAINING expected values 
        title: str None by default
            If title is not None, it is because we must correct a test with random selection of questions
            of type exam_show, of title title
        threshold: float
            threshold=0 thresholds the marks for each question to zero (otherwise negative mark possible)
        weights: dict
            Weight matrix 
            ex: default_weights = {
                (True, True): 1, # True Positive
                (True, False): -1, # False Positive 
                (False, True): 0, # False Negative (forget)
                (False, False): 0 # True Negative
            }
        scale: dict
            Scale = weight of the different questions in the quiz. If no scale, all 
            questions have the same weight for calculating the grade. If weight of a question 
            not specified, it is 1 by default. ex: scale = {'quiz3':4, 'quiz55':0} (all others 
            with a weight of 1).
        maxtries: int
            Number of attempts allowed
        
    Returns
    -------
        Res: pandas dataframe
            rows: students, columns: each quiz + a Note column 
            
    """
    from labquiz import QuizLab
    from labquiz.putils import readData, getAllStudentsAnsvers, getExamQuestions, correctAll

    ## 2 - identification of participants and extraction of responses   
    if title is None:
        students = sorted(list(data_filt["student"].dropna().unique()))
        exam_questions = None
    else:
        exam_questions = getExamQuestions(title, data)
        students = exam_questions.keys()
    
    students_answers = getAllStudentsAnsvers(students, data, maxtries=maxtries) 


    ## 4 - Correct all!
    ResDataFilt = correctAll(students_answers, quiz, data_filt, seuil=seuil, 
                    exam_questions=exam_questions, weights=weights, bareme=bareme, maxtries=maxtries)
    
    if bareme is not None:
        colsToDrop = [quiz_id for quiz_id, val in bareme.items() if val==0 and quiz_id in ResDataFilt.columns]
        return ResDataFilt.drop(columns=colsToDrop)
    return ResDataFilt


def getAllStudentsAnsvers(students, data, maxtries=3):
    students_answers = {student:{} for student in students }
    data_filt =  data.query("event_type in ['validate', 'validate_exam', 'correction'] ")
    #
    for student in students:
        quizzes_done = data_filt.query(f"student == '{student}' ")["quiz_title"].unique()
        for quiz_id in quizzes_done:
            if quiz_id in ["0", "integrity"]: continue
            given = check_ans(student, quiz_id, data_filt, maxtries=maxtries)
            students_answers[student][quiz_id] = given
    return students_answers   
# example: students_answers = getAllStudentsAnsvers(students, data, maxtries=3)


def getExamQuestions(exam_title, data):
    """
    Get from `data` the specific questions asked for each students in the random quiz `exam_title`
    """
    import json
    dq = data.query(f"quiz_title == '{exam_title}' and event_type== 'starting'")
    dqq = dq.drop_duplicates(
       subset=["student"],
     keep="last")
    exam_questions = {}
    for row in dqq.index:
        exam_questions[dqq.loc[row, 'student']] = dqq.loc[row, 'answers']
    return exam_questions
#example
#exam_questions = getExamQuestions("Examen de Truc", data)


def correctAllPrev(students_answers, quiz, df, seuil=0, weights=None, bareme=None, maxtries=3):
    """Correct all answers
    correctAll(students, floors, df, threshold=0, Weight=None, scale=None)
    return Res 
    Corrects all student results from `students` using `sols` solutions

    Arguments
    ---------
        students: dict
            Answers given
        quiz: object 
            QuizLab object into which the topic was loaded
            >> quiz = QuizLab(URL, QUIZFILE, needAuthentification=False, retries=100)
        df: pandas dataframe
            Input table
        threshold: float
            threshold=0 thresholds the marks for each question to zero (otherwise negative mark possible)
        weights: dict
            Weight matrix 
            ex: default_weights = {
                (True, True): 1, # True Positive
                (True, False): -1, # False Positive 
                (False, True): 0, # False Negative (forget)
                (False, False): 0 # True Negative
            }
        scale: dict
            Scale = weight of the different questions in the quiz. If no scale, all 
            questions have the same weight for calculating the grade. If weight of a question 
            not specified, it is 1 by default. ex: scale = {'quiz3':4} (all others 
            with a weight of 1).
        maxtries: int
            Number of attempts allowed
        
    Returns
    -------
        Res: pandas dataframe
            rows: students, columns: each quiz + a Note column 

    """

    import pandas as pd
    
    allquizzes = [q for q in quiz.quiz_bank.keys() if q != "title"]
    students = students_answers.keys()
    Res = pd.DataFrame(index=students, columns=allquizzes)
    if bareme is not None: 
        bareme = {q:1 if q not in bareme.keys() else bareme[q] for q in allquizzes }
    else:
        bareme = {q:1 for q in allquizzes}
    for s in students:
        for quiz_id in allquizzes:
            given = students_answers[s].get(quiz_id, {})
            if given == {}: continue
            try:
                score, score_max = correct_ans(quiz, quiz_id, given, weights=weights)
            except:
                print(f"Erreur pour la correction de {s}, quiz_id={quiz_id}")
            Res.loc[s, quiz_id] = score/score_max
    Res = Res.infer_objects(copy=False).fillna(0)
    if seuil==0: Res[Res < 0] = 0
    poids = pd.Series(bareme)
    Res['FinalMark'] = Res[poids.index].dot(poids) / poids.sum()*20
      
    return Res
"""
students = sorted(list(data["student"].dropna().unique()))
students = [s.title().strip() for s in students]
students_answers = getAllStudentsAnsvers(students, data, maxtries=3)

import time
tic = time.perf_counter()
ResDataFiltBis = correctAllBis(students_answers, quiz, 
                            data_filt, bareme={ 'quiz13':0, 'quiz53':0, 'quiz54':0,
                                               'quiz55':0, 'quiz56':0, 'quiz57':0})
toc = time.perf_counter()
print(f"Temps d'exécution : {toc-tic:.3f} seconde(s)")
"""


#=====================================================
# Security
#
# --------
# Detects if startup parameters
# or during execution have been modified. Detects
# if the source has been modified or monkey patched.
#
#=====================================================



def start_integrity(starting_values, df):
    """Checks that startup settings have not been changed (the name of the student is not necessarily known at this level)

    start_integrity(tarting_values, df)
    return bool


    Settings
    ----------

        starting_values: dict
            Dictionary of keys and values to test for possible modification
        df: pandas dataframe
            Data table
    Returns
    -------
        bool

    """
    # Checks that startup parameters have not been changed
    #dfs = df.query(f"student == '{s}' ")
    allans = df.query(f"event_type == 'starting'")[['notebook_id','parameters','answers']]
    for idx in allans.index:
        params = allans.loc[idx, 'parameters']
        id = allans.loc[idx, 'notebook_id']
        #ans = parse_custom_dict(ans)
        subans = {key:params.get(key,'') for key in starting_values} #sous ensemble de ans avec les bonnes clés
        if not subans == starting_values:
            for key in starting_values:
                if subans[key] != starting_values[key]: 
                    print(f"{id} - enregistrement {idx} : Clé originale '{key}' modifiée de {starting_values[key]} vers {subans[key]}")
            #return False
    return True


        
def check_start_integrity_all_std(starting_values, df):
    """integrity check for all students

    check_start_integrity_all_std(starting_values, df)

    Settings
    ----------
        starting_values: dict
            Dictionary of keys and values to test for possible modification
        df: pandas dataframe
            Data table
    Returns
    -------
        nothing
    """

    start_integrity(starting_values, df)

########        
        

def check_integrity_msg(s, parameters, df):
    """Checks that settings were not changed during execution for student s

    check_integrity(s, parameters, df)
    return bool
    For student `s`, check that
    - machine_id does not change
    - the operating parameters (contained in check_integrity) have not been modified
    Shows if machine modification, parameters modified (and which ones)   

    Settings
    ----------
        s:str
            Student name
        parameters: dict
            Dictionary of keys and values to test for possible modification
        df: pandas dataframe
            Data table
    Returns
    -------
        bool

    """
    out = True
    msg = []
    dfs = df.query(f"student == '{s}' ")
    # - machine_id ne change pas
    if not (len(dfs["notebook_id"].unique()) == 1):
        msg.append(f"{s}: Machine modification for the same student name")
        msg.append(str(dfs["notebook_id"].unique()))
        out = False

    # verify that the check_integrity parameters have not been modified
    recorded_param = dfs['parameters']
    if any( pd.isnull(recorded_param.loc[idx]) for idx in recorded_param.index):
        # t is that they were not saved in parameters but in answers
        recorded_param = dfs.query(f"quiz_title == 'integrity' & event_type == 'check_integrity'")['answers']
    if len(recorded_param)==0: 
        return out,  "\n".join(msg) #No check for this student
    for idx in recorded_param.index: 
        #ans = parse_custom_dict(ans.values[0])
        subrecord = {key:recorded_param.loc[idx].get(key, '') for key in parameters}
        if not subrecord == parameters:
            msg.append(f"{s} - record {idx} :")
            for key in parameters:
                if subrecord[key] != parameters[key]:  
                    msg.append(f"      - Original key '{key}' modified from {parameters[key]} to {subrecord[key]}")
            out = False  
    return out, "\n".join(msg)


def check_integrity(s, parameters, df):
    """Checks that settings were not changed during execution for student s

    check_integrity(s, parameters, df)
    return bool
    For student `s`, check that
    - machine_id does not change
    - the operating parameters (contained in check_integrity) have not been modified
    Shows if machine modification, parameters modified (and which ones)   

    Settings
    ----------
        s:str
            Student name
        parameters: dict
            Dictionary of keys and values to test for possible modification
        df: pandas dataframe
            Data table
    Returns
    -------
        bool 
    """
    out = True
    dfs = df.query(f"student == '{s}' ")
    # - machine_id does not change
    if not (len(dfs["notebook_id"].unique()) == 1):
        print(f"{s}: Machine modification for the same student name")
        print(dfs["notebook_id"].unique())
        out = False

    # verify that the check_integrity parameters have not been modified
    recorded_param = dfs['parameters']
    if any( pd.isnull(recorded_param.loc[idx]) for idx in recorded_param.index):
        # It is that they were not saved in parameters but in answers
        recorded_param = dfs.query(f"quiz_title == 'integrity' & event_type == 'check_integrity'")['answers']
    if len(recorded_param)==0: 
        return True #No check for this student
    for idx in recorded_param.index: 
        #ans = parse_custom_dict(ans.values[0])
        subrecord = {key:recorded_param.loc[idx].get(key, '') for key in parameters}
        if not subrecord == parameters:
            print(f"{s} - enregistrement {idx} :")
            for key in parameters:
                if subrecord[key] != parameters[key]:  
                    print(f"      - Original key '{key}' modified from {parameters[key]} to {subrecord[key]}")
            out = False  
    return out

def check_integrity_all_std(parameters, students, df):  
    """
    Integrity check for all students in students

    check_start_integrity_all_std(starting_values, students, df)

    Settings
    ----------
        starting_values: dict
            Dictionary of keys and values to test for possible modification
        students: list
            List of str (student names)
        df: pandas dataframe
            Data table
    Returns
    -------
        nothing
    """
    for s in students:
        check_integrity(s, parameters, df)

def check_machine(df):
    """Detects if the same machine has been used for multiple student names
    
    check_machine(df)
    
    Parameters
    ----------
        df:       pandas dataframe
    Returns:
    -------
        nothing (Alert for cases)
    """
    
    notebookids = df["notebook_id"].unique()
    for nb in notebookids:
        dfs = df.query(f"notebook_id == '{nb}' ")             
        if not (len(dfs["student"].unique()) == 1):
            print(f"Same machine {nb} used by several students")
            print(dfs['student'].unique())

def machines_for_std(s, df):
    """Returns machine number(s) for a student s

    machines_for_std(s, df)

    Parameters
    ----------
        s:str
            Student Name
        df: pandas dataframe
            Data table
    Returns:
    -------
        list of machines_id for `s`

    """
    dfs = df.query(f"student == '{s}' ")
    return dfs['notebook_id'].unique()       
    
def stds_for_machine(id, df):
    """Returns student names for a machine number

    stds_for_machine(id, df)

    Parameters
    ----------
        id: str
            Machine ID
        df: pandas dataframe
            Data table
    Returns:
    -------
        list of students having used machines_id

    """
    dfs = df.query(f"notebook_id == '{id}' ")
    return dfs['student'].unique()

def check_hash_integrity(df, kind_hash='full', wanted_hash=""):
    
    """Integrity test (source modification, monkey patching, monitored parameters)

    check_hash_integrity(data, kind_hash='full', wanted_hash="")
    return: nothing

    Parameters 
    ----------
        kind_hash: str
            Hash type (`src` or `full`) 
            src for sources, machine independent and `full` depends on src and parameters
        df: pandas dataframe
            Data table
        wanted_hash: str
            If testing the source hash, the reference is `wanted_hash`
    Returns:
    -------
        nothing (Alert for cases)

    """
    
    # test kind_hash: src ou full
    #dfnew = df.query(f"quiz_title == 'integrity' & event_type == 'check_integrity'")[['timestamp', 'notebook_id', 'student', 'parameters', 'answers']]
    dfnew = df[['timestamp', 'notebook_id', 'student', 'parameters', 'answers']].copy()
    
    def get_hash(row, kind_hash):
        params = row.get("parameters")
        if isinstance(params, dict) and kind_hash in params:
            return params[kind_hash]
        answers = row.get("answers") #fallback because before the hash was stored in answers 
        if isinstance(answers, dict):
            return answers.get(kind_hash, "")
        return ""
  
    dfnew.loc[:,"src_hash"] = dfnew.apply(lambda row: get_hash(row, 'src_hash'), axis=1)
    dfnew.loc[:,"full_hash"] = dfnew.apply(lambda row: get_hash(row, 'full_hash'), axis=1)
    
    ids = dfnew.groupby(['student', "notebook_id"])
    colhash='full_hash' if (kind_hash == 'full') else 'src_hash'
    for n, group in enumerate(ids):
        student = group[0][0]
        machine_id = group[0][1]
        g = group[1][colhash]
        if (len(g.unique()) > 1): 
            s = []
            for h in g.unique():
                s.append(dfnew.query(f"{colhash} == '{h}'")['student'].iloc[-1])
              
            if kind_hash=='src': print(f"Source was changed for {s}, machine id {id}")
            if kind_hash=='full': 
                print(f"⚠️ {student}, machine id {machine_id}:")
                print(f"    👉🏼 Source or settings have been changed or monkey patched")
            print("Observed hashes:")
            
            idx = g[g != g.shift()]
            for n,h in enumerate(g.unique()):
                msg = f"⚠️ index [{idx.index[n]}]" if h != wanted_hash else "👍"
                print(h, msg)
            print()
        else:
            if (g.unique() != wanted_hash):
                print(f"⚠️ {student}, machine id {machine_id} :")
                print(f"    👉🏼 Source or settings have been changed or monkey patched")
                print(f"⚠️ index {g.index[0]} hash: {g.loc[g.index[0]]}")

def check_hash_integrity_msg(df, kind_hash='full', wanted_hash=""):
    
    """Integrity test (source modification, monkey patching, monitored parameters)

    check_hash_integrity_msg(data, kind_hash='full', wanted_hash="")
    return: nothing

    Settings
    ----------
        kind_hash: str
            Hash type (`src` or `full`) 
            src for sources, machine independent and `full` depends on src and parameters
        df: pandas dataframe
            Data table
        wanted_hash: str
            If testing the source hash, the reference is `wanted_hash`
    Returns:
    -------
        nothing (Alert for cases)
    """
    
    # test kind_hash: src ou full
    #dfnew = df.query(f"quiz_title == 'integrity' & event_type == 'check_integrity'")[['timestamp', 'notebook_id', 'student', 'parameters', 'answers']]
    dfnew = df[['timestamp', 'notebook_id', 'student', 'parameters', 'answers']].copy()
    
    msg = []
    
    def get_hash(row, kind_hash):
        params = row.get("parameters")
        if isinstance(params, dict) and kind_hash in params:
            return params[kind_hash]
        answers = row.get("answers") #fallback car avant on rangeait le hash dans answers
        if isinstance(answers, dict):
            return answers.get(kind_hash, "")
        return ""
  
    dfnew.loc[:,"src_hash"] = dfnew.apply(lambda row: get_hash(row, 'src_hash'), axis=1)
    dfnew.loc[:,"full_hash"] = dfnew.apply(lambda row: get_hash(row, 'full_hash'), axis=1)
    
    ids = dfnew.groupby(['student', "notebook_id"])
    colhash='full_hash' if (kind_hash == 'full') else 'src_hash'
    for n, group in enumerate(ids):
        student = group[0][0]
        machine_id = group[0][1]
        g = group[1][colhash]
        if (len(g.unique()) > 1): 
            s = []
            for h in g.unique():
                s.append(dfnew.query(f"{colhash} == '{h}'")['student'].iloc[-1])
              
            if kind_hash=='src': print(f"Source code changed for {s}, machine id {id}")
            if kind_hash=='full': 
                msg.append(f"⚠️ {student}, machine id {machine_id} :")
                msg.append(f"    👉🏼 Source or settings have been changed or monkey patched")
            print("hash constatés :")
            
            idx = g[g != g.shift()]
            for n,h in enumerate(g.unique()):
                msg.append(f"{h}: ⚠️ index [{idx.index[n]}]" if h != wanted_hash else "👍")
        else:
            if (g.unique() != wanted_hash):
                msg.append(f"⚠️ {student}, machine id {machine_id} :")
                msg.append(f"    👉🏼 Source or settings have been changed or monkey patched")
                msg.append(f"⚠️ index {g.index[0]} hash: {g.loc[g.index[0]]}")
                
    return '\n'.join(msg)
                
#########

def sha256_bytes(data: bytes) -> str:
    return hashlib.sha256(data).hexdigest()

def fernet_key_from_password(password: str) -> bytes:
    return base64.urlsafe_b64encode(
        hashlib.sha256(password.encode()).digest()
    )

def session_sentinel(pw2: str) -> str:
    return hashlib.sha256(pw2.encode()).hexdigest()


def inject_sentinel(py_path: Path, sentinel: str):
    
    _SENTINEL_RE = re.compile(
    r"^\s*#\s*noqa:\s*E501\s*#\s*.*$",
    re.MULTILINE,
    )
    
    text = py_path.read_text(encoding="utf-8")
    #line = f"# __SESSION_SENTINEL__ = {sentinel}"
    line = f"# noqa: E501  # {sentinel[:16]}"

    if _SENTINEL_RE.search(text):
        text = _SENTINEL_RE.sub(line, text)
    else:
        text = text.rstrip() + "\n" + line + "\n"

    py_path.write_text(text, encoding="utf-8")

    
def hash_distribution(files: list[Path]) -> str:
    h = hashlib.sha256()
    for p in sorted(files):
        h.update(p.read_bytes())
    return h.hexdigest()


def create_secure_tar(
    source_dir,
    output_file,
    password_open,     # PW₁
    password_seal      # PW₂ (secret)
):
    #import io, tarfile
    
    source_dir = Path(source_dir)
    output_file = Path(output_file)
    

    # --- Inserting a sentinel into source files ---
    
    SENTINEL = session_sentinel(password_seal)

    for py in source_dir.rglob("*.py"):
        inject_sentinel(py, SENTINEL)

        
    # --- création du tar  ---
    buf = io.BytesIO()
    with tarfile.open(fileobj=buf, mode="w") as tar:
        tar.add(source_dir, arcname=".")
    tar_bytes = buf.getvalue()

    global_hash = sha256_bytes(tar_bytes) # --- hash du tar ---

    # --- chiffrement ---
    key = fernet_key_from_password(password_open)
    encrypted = Fernet(key).encrypt(tar_bytes)
    output_file.write_bytes(encrypted)

    # 
    src_hash = hash_distribution(list(source_dir.rglob("*.py")))

    print("✔ Archive created:", output_file)
    print("🔐 Global archive integrity hash (to keep):")
    print(global_hash)
    print("🔐 Global hash of source and data files (to be kept) ")
    print(src_hash)

    return global_hash, src_hash

### Check integrity during exam 
# Archive preparation
# Example : 
"""global_hash, src_hash = create_secure_tar(
    source_dir="quiz_data",
    output_file="quiz.tar.enc",
    password_open="mot_pour_ouvrir",
    password_seal="secret_enseignant22"
)"""
#
# ----- Difference between dicts (recursive) -------------
def diff_dicts(
    row_idx,
    current,
    reference,
    ignore_keys=None,
    ignore_paths=None,
    path=""
):
    # Return:
    # anomaly: bool
    # anomalies : str - message describing found anomalies
    # out: dict - dict of anomalies {key:(reference_value, new_value) ...}
    
    ignore_keys = ignore_keys or set()
    ignore_paths = ignore_paths or set()
    anomaly = False
    anomalies = []
    out = {}

    all_keys = set(reference) | set(reference)

    for key in all_keys:
        if key in ignore_keys:
            continue

        cur_path = f"{path}.{key}" if path else key
        if cur_path in ignore_paths:
            continue

        v_cur = current.get(key, "<absente>")
        v_ref = reference.get(key, "<absente>")

        # Nested dictionaries cases
        if isinstance(v_cur, dict) and isinstance(v_ref, dict):
            anomalies.extend(
                diff_dicts(
                    row_idx,
                    v_cur,
                    v_ref,
                    ignore_keys,
                    ignore_paths,
                    cur_path
                )
            )
        else:
            if v_cur != v_ref:
                anomalies.append(
                    f"Ligne {row_idx} - Anomaly, for key '{cur_path}', "
                    f"Initial value {v_ref} modified into {v_cur}"
                )
                anomaly = True
                out[cur_path] = (v_ref, v_cur)

    return anomaly, anomalies, out

def make_ano_report(out, includeRAS=True):
    if len(out) > 0:
        text = "⚠️" 
    else: 
        text = "✅ RAS"
        return text
    for k,v in out.items():
        if k=='full_hash':
            v0, v1 = str(v[0]), str(v[1])
            hash1 = v0[0:5] + '...' + v0[-3:]
            hash2 = v1[0:5] + '...' + v1[-3:]
            text += f"\n key {k} modified from {hash1} to {hash2}"
        else:
            text += f"\n key {k} modified from {v[0]} to {v[1]}"
    return text

def make_anomalies_df_report(df, reference, ignore_keys=[], includeRAS=True):
    # Generate the dataframe report
    messages = []
    Out = {}

    for idx, params in sorted(df["parameters"].items()):
        if isinstance(params, dict) and len(params) != 0:
            dd = diff_dicts(
                    row_idx=idx,
                    current=params,
                    reference=reference,
                    ignore_keys=ignore_keys,
                    ignore_paths=[]
                )
            if dd[0]:
                messages.extend(dd[1])
                Out[idx] = dd[2]
            else:
                if includeRAS:
                    Out[idx] = '' 
                else:
                    pass
                #print(f"pas de souci pour {idx}")
        else:
            pass
            #print(f'idx {idx} pas de dictionnaire')

    Result = pd.DataFrame(columns=['idx', 'timestamp', 'student', 'anomalies'])
    for idx, out in Out.items():
        if not pd.isnull(df.loc[idx, 'student']):
            row = idx, df.loc[idx, 'timestamp'], df.loc[idx, 'student'], make_ano_report(out, includeRAS)
            Result.loc[len(Result)] = row

    Result['timestamp'] = pd.to_datetime(Result["timestamp"].str.replace(r"\s*\(.*\)$", "", regex=True), errors="coerce" )
    Result['timestamp'] = Result['timestamp'].dt.strftime("%Y-%m-%d %H:%M")
    
    return Result

def group_anomalies_per_student(Result):
    df2 = Result.copy()
    if not df2.empty:
        indent = ' '*len(df2.loc[df2.index[0], "timestamp"] + ":")
    else: 
        indent = ""
    df2["entry"] = df2["timestamp"] + ":" + \
df2["anomalies"].str.replace('\n', f"\n {indent}- ")
    #.str.strip('\n').str.replace('\n', f"\n {indent}")

    Result_grouped = (
        df2.sort_values("timestamp")
           .groupby("student", as_index=False)
           .agg(texte=("entry", "\n".join))
    )
    return Result_grouped


import gettext
from pathlib import Path
import os
import locale

'''LOCALE_DIR = Path(__file__).parent / "locales"

_t = gettext.translation("labquiz", 
                         localedir=str(LOCALE_DIR), 
                         fallback=True)

_ = _t.gettext'''

import gettext
from pathlib import Path

# On définit le chemin des locales par rapport à ce fichier
LOCALE_DIR = Path(__file__).parent / "locales"

# Variable interne qui contient l'objet de traduction actuel
_current_trans = gettext.NullTranslations()

'''def set_language(lang_code):
    """Update the global translation object for the whole module."""
    global _current_trans
    _current_trans = gettext.translation(
        "messages", 
        localedir=LOCALE_DIR, 
        languages=[lang_code], 
        fallback=True
    )'''

def get_available_languages():
    """Return a list of available languages."""
    return [
        d.name for d in LOCALE_DIR.iterdir() 
        if d.is_dir() and any(d.rglob("*.mo"))
    ]

def get_best_language():
    # 1. Check environment variables (robust method)
    for env_var in ('LANGUAGE', 'LC_ALL', 'LC_MESSAGES', 'LANG'):
        lang = os.environ.get(env_var)
        if lang:
            # Nettoyage : 'fr_FR.UTF-8' -> 'fr'
            return lang.split('_')[0].split('.')[0].lower()

    # 2. If nothing is found, use the locale module
    try:
        lang, _ = locale.getdefaultlocale()
        if lang:
            return lang.split('_')[0].lower()
    except:
        pass

    return "en" # Fallback

def set_language(lang_code):
    global _current_trans
    available = get_available_languages()
    
    if lang_code not in available and lang_code != "en":
        print(f"Warning: Language '{lang_code}' not found. Available: {available}")
        print("Defaulting to 'en'.")

    _current_trans = gettext.translation(
        "labquiz",
        localedir=LOCALE_DIR, 
        languages=[lang_code], 
        fallback=True
    )


def _(message):
    """Dynamic wrapper that always uses the current language."""
    return _current_trans.gettext(message)

def auto_init_lang():
    detected = get_best_language()
    available = get_available_languages()
    if detected in available:
        set_language(detected)
    else:
        # Si la langue détectée n'est pas traduite, on met l'anglais
        set_language("en")

# Automatic initialization when loading the module
auto_init_lang()



# %load LabQuiz/utils.py
import yaml, json, base64, sys
import ipywidgets as widgets
from IPython.display import display, Markdown
import hashlib, inspect, re, types
from importlib.metadata import metadata
from pathlib import Path
import numpy as np
import pandas as pd
import inspect
from .i18n import _

from types import ModuleType
from typing import Iterable, Tuple




def b64_encode(text: str) -> str:
    return base64.b64encode(text.encode("utf-8")).decode("ascii")

def b64_decode(text: str) -> str:
    return base64.b64decode(text.encode("ascii")).decode("utf-8")


def decode_dict_base64_(s: str) -> dict:
    json_str = base64.b64decode(s[0].encode("ascii")).decode("utf-8")
    return json.loads(json_str)

def decode_dict_base64(encoded_list):
    decoded = []
    for elt in encoded_list:
        json_str = base64.b64decode(elt.encode("ascii")).decode("utf-8")
        decoded.append(json.loads(json_str))
    return decoded

def is_base64_encoded_text(s: str) -> bool:
    try:
        decoded = base64.b64decode(s, validate=True).decode("utf-8")
    except Exception:
        return False

    # Heuristique : texte imprimable
    printable_ratio = sum(c.isprintable() for c in decoded) / len(decoded)
    return printable_ratio > 0.95


def get_macaddress():
    import uuid
    mac = uuid.getnode()
    mac_str = ':'.join(f'{(mac >> ele) & 0xff:02x}' for ele in range(40, -1, -8))
    if (mac >> 40) % 2: mac_str = "00:00:00:00:00:00" #“Probably generated (randomized) MAC address”
    return mac_str

def getUser():
    import os, uuid
    WE_ARE_IN_JUPYTERLITE = "pyodide" in sys.modules or "piplite" in sys.modules
    if WE_ARE_IN_JUPYTERLITE:
        try:
            id_file = ".labquiz_user_id"
            if os.path.exists(id_file):
                # The existing ID is read
                with open(id_file, "r") as f:
                    user_id = f.read().strip()
            else:
                # A new unique ID is generated
                user_id = str(uuid.uuid4())
                with open(id_file, "w") as f:
                    f.write(user_id)
                # Note: In JupyterLite, the file system is 
                # # automatically synced with browser storage.                
                # # Note : Dans JupyterLite, le système de fichiers est 
                # automatiquement synchronisé avec le stockage du navigateur.
        except Exception as e:
            user_id = "erreurUser"            
        return user_id
    else:   # We are not in JupyterLite
        import os
        try:
            user = os.getlogin()
        except:
            user = "erreurUser"
    return user
    
def compute_machine_id():
    import platform
    import hashlib
    raw = "|".join([
        platform.node(),
        platform.system(),
        platform.release(),
        platform.machine(),
        platform.processor(),
        #platform.python_version(),
        getUser(),
        #get_macaddress(),
    ])
    return hashlib.sha256(raw.encode()).hexdigest()[:16]


import ipywidgets as widgets
from IPython.display import display, Markdown

import ipywidgets as widgets
from IPython.display import display, Markdown


class StudentForm:
    def __init__(self):
        self.name = None
        self.output = widgets.Output()

        self.student_lastname = widgets.Text(
        placeholder=_("FAMILY NAME"),
        description=_("FAMILY NAME:"),
        style={'description_width': '70px'},
        layout=widgets.Layout(width="250px"),
        continuous_update=True
    )   
        self.student_firstname = widgets.Text(
        placeholder=_("First name"),
        description=_("First name:"),
        style={'description_width': '70px'},
        layout=widgets.Layout(width="250px"),
        continuous_update=True
    )
        self.save_name_button = widgets.Button(
        description=_("Save"),
        button_style="info",
        icon="check"
    )
    
        self.student_firstname.observe(self.validate, "value")
        self.student_lastname.observe(self.validate, "value")
        
        self.save_name_button.on_click(self.on_save)
        self.student_firstname.on_submit(self.on_submit) 
        self.student_lastname.on_submit(self.on_submit)
        
        self.validate(None)  # Attend la synchro - disabled au départ


    def on_submit(self, _button):
        self.on_save(None)
    
    
    def validate(self, _button):
        self.save_name_button.disabled = not (
            self.student_firstname.value.strip()
            and self.student_lastname.value.strip()
        )

    def on_save(self, _button):
        self.name = (
            self.student_lastname.value.strip().upper()
            + " "
            + self.student_firstname.value.strip().title()
        )
        if len(self.name.strip()) <= 1:
            import time
            time.sleep(0.3)
            self.onsave()
        
        with self.output:
            self.output.clear_output()
            display(Markdown(_("✔️ **Name recorded**") + f" `{self.name}`"))

    
    def display(self):
        enter_msg = _("Enter your first name and LAST NAME here.")
        form = widgets.VBox([
            widgets.HTML("<h2 style='margin-top: 0; margin-bottom: 0; line-height: 1.2;'> {enter_msg} </h2>".format(enter_msg=enter_msg)),
            #widgets.HTML("<b> &nbsp;&nbsp;&nbsp; ⚠️ (appuyez sur Entrée pour valider) </b>"),
            widgets.HBox([self.student_firstname, self.student_lastname]),
            self.save_name_button,
            self.output
           ])
        display(form)

###


def sanitize_dict(d):
    out = {}
    if any( inspect.ismodule(v) for v in d.values() ):
        out = {"modules":{}}

    for k, v in d.items():

        # Modules
        if isinstance(v, types.ModuleType):
            out["modules"][k] = v.__name__

        # NumPy arrays
        elif isinstance(v, np.ndarray):
            out[k] = v.tolist()

        # NumPy scalars
        elif isinstance(v, np.generic):
            out[k] = v.item()

        # Pandas
        elif isinstance(v, (pd.DataFrame, pd.Series)):
            out[k] = v.to_dict() if hasattr(v, "to_dict") else v.tolist()

        # Other non serializable objects → string. // Autres objets non sérialisables → string
        elif hasattr(v, "__dict__"):
            out[k] = str(v)

        #  JSON native types
        else:
            out[k] = v

    return out



###
def get_source_integrity_hash(cls):
    """
    Computes a hash based on the source text of the class's methods.
    """
    try:
        # 1. Entire class source code
        source = inspect.getsource(cls)
        
        # 2. Cleanup for more stability: removing comments, empty lines

        source = re.sub(r'#.*', '', source) 
        source = "\n".join([line.strip() for line in source.splitlines() if line.strip()])
        
        # 3. Hashing
        return hashlib.sha256(source.encode('utf-8')).hexdigest()
    
    except (OSError, TypeError):
            return "Source non disponible"



def get_ultra_integrity_hash(cls):
    """
    Calculates a hash based on bytecode AND constants (fixed values) 
    methods defined locally in the class.
    """
    method_hashes = []

    for name, attr in cls.__dict__.items():
        func = None
        if inspect.isfunction(attr):
            func = attr
        elif isinstance(attr, (staticmethod, classmethod)):
            func = attr.__func__

        if func and hasattr(func, "__code__"):
            code_obj = func.__code__
            
            # 1. bytecode
            bytecode = code_obj.co_code
            
            # 2. Constants (hardcoded values: strings, int, etc.) 
            # # We convert to string to be able to hash it easily
            consts = str(code_obj.co_consts).encode('utf-8')
            
            # We combine the name, bytecode and constants
            signature = f"{name}:".encode('utf-8') + bytecode + consts
            method_hashes.append(hashlib.sha256(signature).hexdigest())

    if not method_hashes:
        return None

    method_hashes.sort()
    final_payload = "".join(method_hashes).encode('utf-8')
    return hashlib.sha256(final_payload).hexdigest()



####### Yet another new version of Hash ############

# Hash independent of python version
# broken down into small modules
# - initial source code
# - living object
# - watchl#ist
# dependencies used
# With module source browsing
# bytecodes deleted

# ---------- Basic functions ----------

def hash_fonction(func: callable) -> bytes:
    """
    Hash robuste d'une fonction ou méthode.
    """
    h = hashlib.sha256()

    # Source if available
    try:
        src = inspect.getsource(func)
        h.update(src.encode("utf-8"))
    except (OSError, IOError):
        pass

    """code = func.__code__
    h.update(code.co_code)
    h.update(repr(code.co_consts).encode("utf-8"))
    h.update(repr(code.co_names).encode("utf-8"))"""

    return h.digest()

def hash_fonction_new(func: callable) -> bytes: #WAS #hash_callable_live(func) -> bytes:
    h = hashlib.sha256()

    # Linked method case → retrieve the real function
    func = getattr(func, "__func__", func)

    # Pure Python function
    if hasattr(func, "__code__"):
        code = func.__code__

        h.update(code.co_code)
        h.update(repr(code.co_consts).encode())
        h.update(repr(code.co_names).encode())
        h.update(repr(code.co_varnames).encode())
        h.update(repr(code.co_freevars).encode())
        h.update(repr(code.co_cellvars).encode())
        h.update(str(code.co_argcount).encode())
        h.update(str(code.co_kwonlyargcount).encode())
        h.update(str(code.co_flags).encode())

        # localization runtime (IPython use it)
        h.update(code.co_filename.encode())
        h.update(str(code.co_firstlineno).encode())

    else:
        # builtin / C-extension
        h.update(repr(func).encode())

    return h.digest()


# ---------- First level functions / Fonctions de premier niveau ----------

def fonctions_premier_niveau(module: ModuleType) -> Iterable[Tuple[str, callable]]:
    for name, obj in vars(module).items():
        if inspect.isfunction(obj) and (obj.__module__ == module.__name__ or obj.__module__ == '__main__'):
            yield name, obj


# ---------- Classes ----------

def methodes_de_classe_old(cls) -> Iterable[Tuple[str, callable]]:
    """
    Returns methods defined directly within the class (not inherited).
    """
    for name, obj in cls.__dict__.items():
        # staticmethod / classmethod
        if isinstance(obj, (staticmethod, classmethod)):
            yield name, obj.__func__
        elif inspect.isfunction(obj):
            #print("name", name)
            yield name, obj

def methodes_de_classe(cls):
    """
    Returns methods defined directly within the class (not inherited, not imported).
    """
    for name, obj in cls.__dict__.items():

        # unwrap staticmethod / classmethod
        if isinstance(obj, (staticmethod, classmethod)):
            func = obj.__func__
        elif inspect.isfunction(obj):
            func = obj
        else:
            continue

        if not func.__qualname__.startswith(cls.__name__ + "."):
            continue

        yield name, func


def methodes_de_classe_plus(cls):
    """
    Return methods defined directly in the class (not inherited, not imported).
    """
    for name, obj in cls.__dict__.items():

        # 1. Extraire l'objet fonction réel
        if isinstance(obj, (staticmethod, classmethod)):
            func = obj.__func__
        elif inspect.isfunction(obj):
            func = obj
        else:
            continue

        #2. Eliminate functions coming from another module
        if func.__module__ != cls.__module__:
            continue

        #3. Eliminate non-lexically defined functions in the class
        if not func.__qualname__.startswith(cls.__name__ + "."):
            continue
        
        yield name, func


def hash_classe(cls) -> bytes:
    """
    Hash of a class based on its methods and its name.
    """
    h = hashlib.sha256()

    # Name + bases
    h.update(cls.__name__.encode("utf-8"))
    for base in cls.__bases__:
        h.update(base.__name__.encode("utf-8"))

    # Methods
    for name, method in sorted(methodes_de_classe(cls), key=lambda x: x[0]):
        h.update(name.encode("utf-8"))
        h.update(hash_fonction(method))
        #print(name, "-->", h.hexdigest())

    return h.digest()


def classes_du_module(module: ModuleType) -> Iterable[Tuple[str, type]]:
    for name, obj in vars(module).items():
        if inspect.isclass(obj) and obj.__module__ == module.__name__:
            yield name, obj


# ---------- Hash global du module ----------

def hash_module(module: ModuleType) -> str:
    h = hashlib.sha256()

    # Fonctions
    for name, func in sorted(fonctions_premier_niveau(module), key=lambda x: x[0]):
        h.update(b"F")
        h.update(name.encode("utf-8"))
        h.update(hash_fonction(func))
        #print(name, "-->", h.hexdigest())

    # Classes
    for name, cls in sorted(classes_du_module(module), key=lambda x: x[0]):
        h.update(b"C")
        h.update(name.encode("utf-8"))
        h.update(hash_classe(cls))

    return h.hexdigest()

def hash_methodes_vivantes(obj) -> str:
    h = hashlib.sha256()

    for name, func in sorted(methodes_vivantes_objet(obj)):
        h.update(b"M")
        h.update(name.encode())
        h.update(hash_fonction(func))
        
    return h.hexdigest()

## This works!

def methodes_vivantes_objet(obj):
    cls = obj.__class__
    module_name = cls.__module__

    for name in dir(obj):
        if name.startswith("__"):
            continue

        try:
            attr = getattr(obj, name)
        except Exception:
            continue

        if not callable(attr):
            continue

        # bound method → actual function
        func = getattr(attr, "__func__", attr)

        # 1. Exclude external imports
        if func.__module__ != module_name:
            continue

        #2. Exclude external inherited methods
        if not func.__qualname__.startswith(cls.__name__ + "."):
            continue

        yield name, func

def hash_classe_vivante(cls) -> bytes:
    h = hashlib.sha256()

    for name in dir(cls):
        if name.startswith("__"):
            continue

        try:
            attr = getattr(cls, name)
        except Exception:
            continue

        if not callable(attr):
            continue

        func = getattr(attr, "__func__", attr)

        if func.__module__ != cls.__module__:
            continue

        if not func.__qualname__.startswith(cls.__name__ + "."):
            continue

        h.update(name.encode())
        h.update(hash_fonction(func))

    return h.digest()

def hash_dependances_modules(obj, modules_cibles) -> bytes:
    """
    Hash of global functions in target_modules used by class methods.

    obj: class whose graph we want to secure
    target_modules: list of module names to include (ex: ["main", "utils"])
    """
    
    cls = obj.__class__
    module_name = cls.__module__
    package = obj.__module__.split('.', 1)[0]
    
    h = hashlib.sha256()
    # extract dependencies used by class methods
    dependances = set()
    for name, func in methodes_vivantes_objet(obj):
        dependances |= set(func.__code__.co_names)
    #print(dependances)
    # browse targeted modules only
    for module in modules_cibles:

        for name in sorted(dependances):
            if not hasattr(module, name):
                continue
            obj = getattr(module, name)
            #print("pre-test", h.hexdigest())
            # ne traiter que les fonctions Python
            if isinstance(obj, (types.FunctionType, types.MethodType)):
                if obj.__module__ != module.__name__:
                    continue
                #h.update(name.encode())
                #h.update(hash_methodes_vivantes(obj).encode())
                h.update(hash_fonction(obj))
                #print("dependances:", name, hash_fonction(obj).hex() )
                #print("  -->  ", h.hexdigest())
            else:
                pass
                # if necessary, we can also hash other objects, for example classes or constants
                #h.update(name.encode())
                #h.update(repr(obj).encode())
     
    return h.hexdigest()


# --------- watchlist --------------------

def stable_value(val):
    if val is None or isinstance(val, (int, float, bool, str)):
        return repr(val)
    if isinstance(val, (list, tuple)):
        return "[" + ",".join(stable_value(v) for v in val) + "]"
    if isinstance(val, dict):
        return "{" + ",".join(
            f"{stable_value(k)}:{stable_value(v)}"
            for k, v in sorted(val.items(), key=lambda x: repr(x[0]))
        ) + "}"
    return f"<UNHASHABLE:{type(val).__name__}>"


def hash_watchlist(obj, WATCHLIST=[]):
    cfg_parts = []
    for name in WATCHLIST:
        if hasattr(obj, name):
            val_str = stable_value(getattr(obj, name))
            cfg_parts.append(f"{name}:{val_str}")
    cfg_parts.sort()
    payload = "|".join(cfg_parts)
    return hashlib.sha256(payload.encode()).hexdigest()

def get_big_integrity_hash(obj, modules=[], WATCHLIST=[]):
    """
    Returns a dict with: 
    - "source_hash": modules_hash, # Sources 
    - "watchlist_hash": watchlist_hash, # monitored parameters 
    - "live_object":living_object, # living object methods -- identifies monkey patching on the object 
    - "dependencies": dependencies, # the living methods of the modules used by the current object 
    - "full_hash": full_hash
    """
    
    modules_hash = []
    package = obj.__module__.split('.', 1)[0]
    modules = [sys.modules[package+"."+mod] for mod in modules]
    for module in modules:
        mh = hash_module(module)
        modules_hash.append(mh)
    
    watchlist_hash = hash_watchlist(obj, WATCHLIST)
    living_object = hash_methodes_vivantes(obj) 
    dependances = hash_dependances_modules(obj, modules)
    
    payload_parts = modules_hash.copy()
    payload_parts.append(watchlist_hash)
    payload_parts.append(living_object)
    payload_parts.append(dependances)
    
    # Combinaison
    payload = "|".join(payload_parts)
    full_hash = hashlib.sha256(payload.encode()).hexdigest()

    return {
        "source_hash": modules_hash,
        "watchlist_hash": watchlist_hash,
        "live_object":living_object,
        "dependances":dependances,
        "full_hash": full_hash
    }

def get_full_object_hash(obj, modules=[], WATCHLIST=[]):
    """
    Returns a full_hash, combining all hashes
    - "full_hash": full_hash
    """
    
    modules_hash = []
    package = obj.__module__.split('.', 1)[0]
    modules = [sys.modules[package+"."+mod] for mod in modules]
    """for module in modules:
        mh = hash_module(module)
        modules_hash.append(mh)"""
    
    watchlist_hash = hash_watchlist(obj, WATCHLIST)
    living_object = hash_methodes_vivantes(obj) 
    #dependances = hash_dependances_modules(obj, modules)
    
    payload_parts = []
    #payload_parts = modules_hash.copy()
    payload_parts.append(watchlist_hash)
    payload_parts.append(living_object)
    #payload_parts.append(dependances)
    
    # Combinaison
    payload = "|".join(payload_parts)
    full_hash = hashlib.sha256(payload.encode()).hexdigest()

    return full_hash

# examples:
# get_big_integrity_hash(quiz, modules=["main", "utils"])
# get_full_object_hash(quiz_dev, modules=["main", "utils"], WATCHLIST=['retries'])


def prep(n, quiz_id, quizfile, internetOK=False, p=None):
    if p is None: p=n 
    if n > 0:
        h = prep(n - 1, quiz_id, quizfile, internetOK, p)
    else:
        stack = inspect.stack() 
        context = [frame.function for frame in stack[2:7]]
        nc = None
        frame = inspect.stack()[p+2].frame

        if 'self' in frame.f_locals:
            nc = frame.f_locals['self'].__class__.__name__
        elif 'cls' in frame.f_locals:
            nc = frame.f_locals['cls'].__name__
        context.extend([nc, quizfile])
        while len(context) < 6: context.append("none")
        context.extend([internetOK, quiz_id])

        fingerprint = "-".join(context).encode()
        hash_digest = hashlib.sha256(fingerprint).digest()
        h = base64.urlsafe_b64encode(hash_digest)
    return h


def is_encrypted(s):
    from cryptography.fernet import Fernet, InvalidToken
    h = b'zzFJ0WBRXuFC1qJOFGxeqt3NUNcFR9vdZqFkCry-DYw='
    f = Fernet(h)
    try:
        dec = f.decrypt(s).decode('utf-8')
        return True
    except: # InvalidToken:
        return False

def calculate_quiz_score(quiz_type, user_answers, propositions, weights=None, constraints=None):
    """ 
    constraints: List of dicts ex: [{"indexes": (0, 1), "type": "XOR", "malus": 2}] 
    - XOR (Exclusion) A and B must be different 
    - IMPLY (Implication) If A is TRUE then B is TRUE 
    - SAME (Coherence) A and B are equivalent (same value) 
    - IMPLYFALSE: If A is TRUE, then B MUST be FALSE 
    """ 
    # Reference matrix (If nothing is defined in the proposal) and if 
    # another reference matrix is not passed as an argument 
    # Format: (User_Response, Expected)
    
    # Precaution - Change of key name in new YAML structure. Check and rename if old version
    for prop in propositions: 
        if "reponse" in prop:
            prop["answer"] = prop.pop("reponse")
    if quiz_type == "qcm": quiz_type = "mcq"
    if quiz_type == "qcm-template": quiz_type = "mcq-template"

    # Weights
    if weights is None:
        default_weights = {
                (True, True):   1,  # Vrai Positif
                (True, False): -1,  # Faux Positif 
                (False, True):  0,  # Faux Négatif (oubli)
                (False, False): 0   # Vrai Négatif
            }
    else:
        default_weights = weights

    TYPE_MAP = {
        "int": int,
        "bool": lambda s: s.lower() == "true" if isinstance(s, str) else s,
        "float": float,
               }
    
    score = 0.0
    total_possible = 0.0
    if not isinstance(user_answers, dict): 
        print(_("❌ Error : user_answers must be a dictionary."))

    # Correction in the case of a template
    if quiz_type in ['numeric-template', 'mcq-template']:
        context = user_answers.get('context', {})
        #print(context)
        for p in propositions:
            pexpect =  p["expected"]
            ptype = p.get("type", bool if quiz_type == "mcq" else float)
            ptype = TYPE_MAP.get(ptype, ptype)
            if "modules" in context:  #we import the modules necessary to calculate the answer - HOT
                toexec = ""
                for k,v in context["modules"].items():
                    toexec += f"import {v} as {k}\n"
                toexec += "result="+pexpect
                exec(toexec, context)
                p["expected"] = ptype(context["result"])
            else:    
                p["expected"] = ptype(eval(pexpect,{},context)) if isinstance(pexpect, str) else pexpect
            p['proposition'] = p['proposition'].format(**context)
        quiz_type = quiz_type.split('-')[0]
        user_answers.pop("context", None)
        
    for answer, prop in zip(user_answers.values(), propositions):
        expected = prop.get("expected", None)
        # default was False
        if expected is None: return 0, 1 #No solution known in proposition
        user_val = bool(answer)
        case = (user_val, expected)
        
        if quiz_type in ["mcq", "mcq-template"]:
            #1. Calculation of the theoretical total
            total_possible += prop.get("bonus", default_weights[(True, True)]) if expected \
                   else prop.get("bonus", default_weights[(False, False)])

            #2. Calculation of the score for this proposition
            if user_val == expected:
                # Correct Case (VP ou VN)
                val = prop.get("bonus", default_weights[case])
                score += val
            else:
                # Incorrect Case (FP or FN)
                val = prop.get("malus", default_weights[case])              
                # We make sure that the penalty is indeed deducted
                score -= abs(val)

        elif quiz_type in ["numeric", "numeric-template"]:
            pexpected = float(prop.get("expected", 0))
            bonus = prop.get("bonus", 1)
            total_possible += bonus
            diff = abs(answer - pexpected)
            tol = max(prop.get("tolerance_abs", 0), 
                      prop.get("tolerance", 0.01) * abs(pexpected)) 
            score += bonus if diff <= tol else -abs(prop.get("malus", 0))

           
    # contraints
    if quiz_type == "mcq" and constraints:
        for rule in constraints:
            idx1, idx2 = rule["indexes"]
            ans1, ans2 = bool(user_answers[idx1]), bool(user_answers[idx2])

            violation = False
            r_type = rule.get("type", "XOR").upper()

            if r_type == "XOR": #The two cannot be identical
                if not (ans1 != ans2) :
                    violation = True
                    #print(f"violation XOR entre {idx1} et {idx2}")
            elif r_type == "IMPLY": # If 1 is TRUE, then 2 MUST be TRUE (otherwise contradiction)
                if ans1 and not ans2:
                    violation = True
                    #print(f"violation IMPLY entre {idx1} et {idx2}")
            elif r_type == "IMPLYFALSE": # If 1 is TRUE, then 2 MUST be FALSE (otherwise contradiction)
                if ans1: 
                    if ans2:
                        violation = True
                    #print(f"violation IMPLY entre {idx1} et {idx2}")
            elif r_type == "SAME": # Must have the same value
                if ans1 != ans2:
                    violation = True
                    #print(f"violation SAME between {idx1} and {idx2}")

            if violation:
                # We subtract the contradiction malus
                score -= abs(rule.get("malus", 1))

    return score, total_possible



def get_package_directory(package_name):
    """ Returns the directory containing the package whose name is given.

    Settings
    ----------
    package_name: str
        Package name

    Returns
    -------
    Path
        Package folder

    Raises
    -------
    ImportError
        If the package is not available.

    # Example of use
    package_dir = get_package_directory("labquiz")
    print(package_dir)  
    """
    import importlib.util
    spec = importlib.util.find_spec(package_name)
    if spec is None:
        raise ImportError(f"Package {package_name} is unavailable.")
    return Path(spec.origin).parent

def get_package_hash(package_name):
    """ Returns the hash of the package whose name is passed.
    Settings
    ----------
    package_name: str
        Package name

    Returns
    -------
    str
        Package hash (or None if the package has no hash)

    Raises
    -------
    ImportError
        If the package is not available.

    Notes
    -----
    The hash is stored in the Keywords field of the METADATA file
    of the package. It starts with "hash:".

    Example of use
    ---------------------
    package_hash = get_package_hash("labquiz")
    print(package_hash)
    """
    from importlib.metadata import metadata
    meta = metadata(package_name)
    #print(meta)
    keywords = meta.get_all('Keywords', [])
    # We are looking for the element that begins with "hash:"
    kz = keywords[0]
    for k in kz.split(','):
        if k.startswith("hash:"):
            return k.split(":", 1)[1]
    return None

def package_hash(package_dir, includeOnly=None, exclude=None, algo="sha256"):
    """
    Compute the hash of a package.

    Parameters
    ----------
    package_dir : Path or str
        Path to the package directory.
    includeOnly : set or list, optional
        List of ONLY file names to include in the hash computation.
    exclude : set or list, optional
        List of file names or directory names to exclude from the hash computation.
    algo : str, optional
        Hash algorithm to use (default is 'sha256').

    Returns
    -------
    tuple
        A tuple containing the hash as a string and the list of file paths that were used to compute the hash.
    """
    
    package_dir = Path(package_dir) if isinstance(package_dir, str) else package_dir
    h = hashlib.new(algo)
    exclude = exclude or set()

    if includeOnly is not None:
        files = sorted(
            p for p in package_dir.rglob("*")
            if p.is_file() and p.name in includeOnly
            )
    else:
        files = sorted(
            p for p in package_dir.rglob("*")
            if p.is_file() and p.name not in exclude and not any(ex in p.parts for ex in exclude)
        )

    for path in files:
        h.update(path.read_bytes())

    return h.hexdigest(), files

def check_installed_package_integrity():
    from importlib.metadata import version, PackageNotFoundError
    import importlib.util
    
    spec = importlib.util.find_spec("labquiz")
    if spec is None:
        print(_("⚠️ LabQuiz is not installed!"))
        return 

    PACKAGE_NAME = __package__.split(".", 1)[0]
    try:
        __version__ = version(PACKAGE_NAME)
    except PackageNotFoundError:
        try:
            __version__  =  version("labquiz") + " (dev)"
        except PackageNotFoundError:
            return "non installé"
        
    EXCLUDE = {"putils.py", "__pycache__", ".ipynb_checkpoints", ".DS_Store"}
    INCLUDEONLY = {"__init__.py", "i18n.py", "main.py", "utils.py"}
    labdir = get_package_directory("labquiz")
    installed_hash, f = package_hash(labdir, includeOnly=INCLUDEONLY, exclude=EXCLUDE)
    recorded_hash = get_package_hash("labquiz")
    if installed_hash != recorded_hash:
        print(_("⚠️ Package hash different than expected -- this is recorded"),""" 
              Installed: {installed_hash} 
              Expected: {recorded_hash}""".format(installed_hash=installed_hash, recorded_hash=recorded_hash))
        return False
    print(f"LabQuiz, {PACKAGE_NAME} version {__version__}")
    return True
#_ = check_installed_package_integrity()

def _sha256_bytes(data: bytes) -> str:
    return hashlib.sha256(data).hexdigest()

def _fernet_key_from_password(password: str) -> bytes:
    return base64.urlsafe_b64encode(
        hashlib.sha256(password.encode()).digest()
    )

def compute_local_hash(base_dir: Path) -> str:
    h = hashlib.sha256()
    for p in sorted(base_dir.rglob("*.py")):
        h.update(p.read_bytes())
    return h.hexdigest()

